# Copyright 2001-2014 Rally Software Development Corp. All Rights Reserved.

require "time"
require "socket"
require "open3"

require "vcseif/connection"       # for VCSConnection
require "vcseif/utils/exceptions" # for OperationalError

############################################################################################

class MercurialConnection < VCSConnection

  HG_CONN_VERSION = "1.2.5"

  HG_LOG_STYLE = %{changeset = "Ident: {rev}-{node}\\nAuthor: {author}\\nCommitted: {date}\\nHumanzDate: {date|rfc3339date}\\nMessage: {desc}\\n{file_adds}{file_mods}{file_dels}\\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\\n"
file_add = "    A {file_add}\\n"
file_mod = "    M {file_mod}\\n"
file_del = "    M {file_del}\\n"
}

  attr_reader :mercurial, :backend_version, :username_required, :password_required
  attr_reader :server, :repository_name, :uri, :debug, :max_items, :localhost

  def initialize(config, logger)
    super(logger)
    @mercurial = nil
    internalizeConfig(config)
    @backend_version = ""
    @username_required = false
    @password_required = false
  end

  def name()
    return "Mercurial"
  end

  def version()
    return HG_CONN_VERSION
  end

  def getBackendVersion()
    %{
          Conform to Connection subclass protocol and provide the version of the system
          we are "connecting" to.
       }
    return @backend_version
  end

  def internalizeConfig(config)
    super(config)
    @server = config['Server'] || Socket.gethostname
    @repository_name = config["RepositoryBase"] || nil
    @uri = config["Uri"] || false
    @debug = config["Debug"] || false
    @max_items = config["MaxItems"] || 1000
    @localhost = _targetServerIsLocalhost(@server)

    if ['false', 'no', 'off', '0', 'disable'].include?(@debug.to_s.downcase)
      @debug = false
    end
  end

  def connect()
    """
          Attempt to run our VCS command to obtain the VCS version.
    """
    @log.info("Connecting to Mercurial")
    begin
      command_vector = base_command_vector
      command_vector << 'hg' << '--version'
      output, errors, status = Open3.capture3(command_vector.join(" "))
      @backend_version = output.split("\n").first
    rescue Exception => ex
      @log.debug(ex.message)
      problem = "Unable to interrogate Mercurial for version information"
      operr = VCSEIF_Exceptions::OperationalError.new(problem)
      raise operr, problem
    end
    if status != 0
      problem = "Unable to connect to Mercurial server at #{@server} to obtain Mercurial software version information using this command: #{command_vector.join(" ")}"
      @log.error(problem)
      @log.error("stderr output:\n"+ errors)
      brief = "Unable to obtain Mercurial server software version information"
      operr = VCSEIF_Exceptions::OperationalError.new(brief)
      raise operr, problem
    end
    @log.info("Connected to Mercurial server: %s %s" % [@server, @backend_version])
    @log.info("RepositoryBase: %s" % @repository_name)
    @log.info("Uri: %s" % @uri)
    @connected = true
    return true
  end

  def disconnect()
    """
          Get rid of any log-styler file we may have written (whether locally or on remote Hg Server)
          and then reset our mercurial instance variable to nil
    """
    _removeLogStyleFile()
    @mercurial = nil
    @connected = false
  end

  def base_command_vector()
    """
          Check to see if @server is localhost or not.
          If @server is localhost, we don't need to bother with ssh
          Otherwise we have to use ssh to run our VCS command on @server.
          (and use ssh keys / assume ssh key access for authentication)
    """
    command_vector = []
    if !@localhost
      command_vector << 'ssh'
      command_vector << '-i' << @sshkey_path unless @sshkey_path.nil?
      server_str = @server
      server_str = "#{@username}@#{@server}" unless @username.nil?
      command_vector << server_str
    end
    return command_vector
  end


  def getRecentChangesets(ref_time)
    """
          Obtain all Changesets created in Mercurial at or after the ref_time parameter
          which is a Time instance
    """
    merc_ref_time = ref_time.to_s
    pending_operation = "Detecting recently added Mercurial Changesets (added on or after %s)"
    @log.info(pending_operation % merc_ref_time)

    changeset_info = runLogCommand(@server, @username, @password,
                                   @repository_name, merc_ref_time)
    if changeset_info.nil? or changeset_info.class.name != "String" or changeset_info.empty?
      @log.error("hg log command did not generate usable changeset information")
      return []
    end

    if @debug
      File.open('mrc.rev.hist', 'w') { |f| f.write(changeset_info) }
    end

    chunks = _changesetChunks(changeset_info)
    changesets = chunks.collect { |cset_info| MercurialChangeset.new(cset_info) }

    log_msg = "  %d recently added Mercurial Changesets detected"
    @log.info(log_msg % changesets.length)
    changesets.each { |cset| @log.debug(cset.details()) }
    if not changesets.empty?
      oldest_changeset = changesets.first
      last_changeset = changesets.last
      if changesets.length > 1
        @log.info("Date of oldest Mercurial commit in chunk: #{oldest_changeset.commit_timestamp}")
      end
      @log.info("Date of latest Mercurial commit in chunk: #{last_changeset.commit_timestamp}")
    end

    return changesets
  end


  private
  def _changesetChunks(payload)
    """
          TODO: do we need to do something less resource intensive for extremely large log outputs?
    """
    chunks = payload.split(/\^{40,}\r?\n/)
    return chunks
  end

  def _writeLogStyleFile()
    %{
          If the repository that is to be interrogated is on the localhost,
          we can write the ".log-styler" file in the current directory.
          However, if the repository that is to be interrogated is on a remote host,
          we'll have to write the ".log-styler" file in /tmp/.log-styler on the
          remote host.
       }
    style_file_name = ".log-styler"

    File.open(style_file_name, 'w') { |f| f.write(HG_LOG_STYLE) }
    if @localhost
      style_file_path = "%s/%s" % [Dir.pwd, style_file_name]
      return style_file_path
    end

    style_file_path = _writeRemoteLogStyleFile(style_file_name)
    return style_file_path
  end

  def _writeRemoteLogStyleFile(style_file_name)
    # we have to write a style file on the remote host;
    # do this by scp'ing the style_file_name it to the remote, opening up the permissions
    # and then remove the local

    command_vector = ["scp"]
    command_vector << "-i" << @sshkey_path unless @sshkey_path.nil?
    command_vector << style_file_name
    dest = "#{@server}"
    if not @username.nil?
      dest = "%s@%s" % [@username, @server]
    end
    style_file_path = "/tmp/%s" % style_file_name
    command_vector << "%s:%s" % [dest, style_file_path]
    begin
      result = "" and IO.popen(command_vector) { |sc| result = sc.read() }
    rescue => ex
      problem = "unable to write /tmp/.log-styler file on remote host: %s" % [@server]
      @log.error(problem)
      operr = VCSEIF_Exceptions::OperationalError.new(problem)
      raise operr, problem
    end

    command_vector = base_command_vector
    command_vector << "chmod" << "777" << style_file_path
    begin
      result = "" and IO.popen(command_vector) { |sc| result = sc.read() }
    rescue => ex
      problem = "unable to chmod 777 /tmp/.log-styler file on remote host: %s" % [@server]
      @log.warn(problem)
    end

    begin
      File.remove(style_file_name)
    rescue Exception => ex
    end

    return style_file_path
  end

  def _removeLogStyleFile()
    %{
          If the repository that is to be interrogated is on the localhost,
          we can remove the ".log-styler" file in the current directory.
          However, if the repository that is to be interrogated is on a remote host,
          we'll have to remove the ".log-styler" file in /tmp/.log-styler on the
          remote host.
       }
    style_file_name = ".log-styler"

    if not @localhost
      status = _removeRemoteLogStyleFile(style_file_name)
    end
    style_file_path = "%s/%s" % [Dir.pwd, style_file_name]
    File.delete(style_file_path) if File.exists?(style_file_path)
  end

  def _removeRemoteLogStyleFile(style_file_name)
    # we have to remove the style file on the remote host;
    command_vector = base_command_vector
    style_file_path = "/tmp/%s" % style_file_name
    command_vector << 'rm' << '-f' << style_file_path
    output, errors, status = Open3.capture3(command_vector.join(" "))
    if not errors.empty?
      # it is possible to fail if the target box has the /tmp directory permissioned as drwxrwxrwt
      # where the t is the sticky designator indicating files in there can _only_ be removed by the
      # file owner
      @log.warn(errors)
    end
    return status == 0
  end

  def runLogCommand(server, user, password, repo, ref_time)
    """
          ref_time comes in as ISO-8601-like format (YYYY-mm-dd HH:MM:SS)
    """
    style_file_path = _writeLogStyleFile()

    #log_cmd = 'hg log --cwd /var/repos/hg/dev/project1 --date ">2011-12-30 09:43:29" --limit <max_items> --style <style_file_name> '
    # a --limit 1000 specifies that only the last 1000 commits will be reported

    ref_time_spec = '">%s"' % ref_time
    ref_time_spec = '\'">%s"\'' % ref_time if !@localhost # employ wacky quote protection technique for ssh'ing

    # for Mercurial, we need to use the --cwd ${repo_dir} to run the hg log command successfully
    log_options = ['--cwd', "#{@repository_name}",
                   '--date', ref_time_spec,
                   '--rev', "0:tip",
                   '--limit', "%d" % @max_items,
                   '--style', style_file_path.count(" ") > 0 ? "\"#{style_file_path}\"" : style_file_path
    ]
    command_vector = base_command_vector
    command_vector << 'hg' << 'log'
    log_options.each { |lopt| command_vector << lopt }
    @log.debug(command_vector.join(" "))
    output, errors, status = Open3.capture3(command_vector.join(" "))
    # TODO: watch out to see if you have to specify some encoding or do some decoding (UTF-8)
    ##
    ##        puts "hg log output:"
    ##        puts "#{output}"
    ##
    exit_code = status.to_s.strip.split(' ').pop
    if exit_code.to_i > 0
      problem = "hg log command error detected: %s, exit code: %s" % [errors, exit_code]
      @log.error(problem)
      operr = VCSEIF_Exceptions::OperationalError.new(problem)
      raise operr, problem
    end

    if @debug
      puts "HG_LOG_STYLE file: %s" % style_file_path
    else
      begin
        Dir.remove(style_file_name)
      rescue Exception => ex
      end
    end

    return output
  end

end

##############################################################################################

class MercurialChangeset

  RALLY_TS_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

  attr_reader :ident, :node, :rev_num, :commit_timestamp, :offset_secs
  attr_accessor :author, :humanz_date
  attr_reader :message
  attr_reader :adds, :mods, :dels

  def initialize(log_text)
    """
          You'll get a multiline block of text that looks like:
             Ident: 9-abc123def0988342acdc345ecacabeef00babe
             Author: Foo Bar <foobar@badcompany.com>
             Committed: 132450324323.019800
             HumanzDate: 2012-03-12T08:36:42-05:00
             Message: DE1073 here is some message text
                 A somefile.name
                 A another file name.txt
                 M this-file-got-changed.doc
                 R and this was removed.ed
    """

    @offset_secs = 0
    @adds = []
    @mods = []
    @dels = []

    _extractFields(log_text)
  end

  private
  def _extractFields(text_block)
    log_lines = text_block.split("\n")
    ident_line = log_lines[0]
    author_line = log_lines[1]
    committed_line = log_lines[2]
    humanz_date_line = log_lines[3]
    message_line = log_lines[4]
    file_lines = log_lines[5..-1]
    ident_info = ident_line.split('Ident: ')[1]
    rev_num, node = ident_info.split('-')
    @ident = node
    @node = node
    @rev_num = rev_num
    @author = author_line.split('Author: ')[1]

    committed_value = committed_line.split("Committed: ")[1]
    commit_epoch_seconds, commit_offset_info = committed_value.split('.')
    @committed = commit_epoch_seconds.to_i
    @commit_timestamp = Time.at(@committed).utc.strftime(RALLY_TS_FORMAT)
    begin
      if commit_offset_info.include?('-')
        @offset_secs = commit_offset_info.split('-')[1].to_i
      elsif commit_offset_info.include?('+')
        @offset_secs = commit_offset_info.split('+')[1].to_i
      else
        @offset_secs = commit_offset_info.to_i
      end
    rescue Exception => ex
      @offset_secs = 0
    end

    @humanz_date = humanz_date_line.split('HumanzDate: ')[1]
    @message = message_line.split('Message: ')[1]

    file_lines.each do |fline|
      next unless fline.start_with?('    ')
      action, file_name = fline.strip().split(/ /, 2)
      @adds << file_name if action == 'A'
      @mods << file_name if action == 'M'
      @dels << file_name if action == 'R'
    end
  end

  public
  def details()
    ident = "Ident: %s-%s" % [@rev_num, @node]
    author = "Author: %s" % @author
    committed = "Committed: %s" % @commit_timestamp
    pepul_date = "ISO8601-Date: %s" % @humanz_date
    message = "Message: %s" % @message
    elements = [ident, author, committed, pepul_date, message]
    adds = @adds.collect { |a| '    A %s' % a }.join("\n") unless @adds.empty?
    mods = @mods.collect { |m| '    M %s' % m }.join("\n") unless @mods.empty?
    dels = @dels.collect { |d| '    D %s' % d }.join("\n") unless @dels.empty?
    elements << adds unless @adds.empty?
    elements << mods unless @mods.empty?
    elements << dels unless @dels.empty?
    elements << " "
    pill = elements.join("\n")
    return pill
  end

end


